/**
  * vee-validate v3.0.0-beta.0
  * (c) 2019 Abdelrahman Awad
  * @license MIT
  */
import Vue from 'vue';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function __awaiter(thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

var isNaN = function (value) {
    // NaN is the one value that does not equal itself.
    return value !== value;
};
/**
 * Checks if the values are either null or undefined.
 */
var isNullOrUndefined = function (value) {
    return value === null || value === undefined;
};
/**
 * Creates the default flags object.
 */
var createFlags = function () { return ({
    untouched: true,
    touched: false,
    dirty: false,
    pristine: true,
    valid: false,
    invalid: false,
    validated: false,
    pending: false,
    required: false,
    changed: false
}); };
/**
 * Shallow object comparison.
 */
var isEqual = function (lhs, rhs) {
    if (lhs instanceof RegExp && rhs instanceof RegExp) {
        return isEqual(lhs.source, rhs.source) && isEqual(lhs.flags, rhs.flags);
    }
    if (Array.isArray(lhs) && Array.isArray(rhs)) {
        if (lhs.length !== rhs.length)
            return false;
        for (var i = 0; i < lhs.length; i++) {
            if (!isEqual(lhs[i], rhs[i])) {
                return false;
            }
        }
        return true;
    }
    // if both are objects, compare each key recursively.
    if (isObject(lhs) && isObject(rhs)) {
        return (Object.keys(lhs).every(function (key) {
            return isEqual(lhs[key], rhs[key]);
        }) &&
            Object.keys(rhs).every(function (key) {
                return isEqual(lhs[key], rhs[key]);
            }));
    }
    if (isNaN(lhs) && isNaN(rhs)) {
        return true;
    }
    return lhs === rhs;
};
/**
 * Gets the value in an object safely.
 */
var getPath = function (path, target, def) {
    if (def === void 0) { def = undefined; }
    if (!path || !target)
        return def;
    var value = target;
    path.split('.').every(function (prop) {
        if (prop in value) {
            value = value[prop];
            return true;
        }
        value = def;
        return false;
    });
    return value;
};
/**
 * Parses a rule string expression.
 */
var parseRule = function (rule) {
    var params = [];
    var name = rule.split(':')[0];
    if (includes(rule, ':')) {
        params = rule
            .split(':')
            .slice(1)
            .join(':')
            .split(',');
    }
    return { name: name, params: params };
};
/**
 * Debounces a function.
 */
var debounce = function (fn, wait, token) {
    if (wait === void 0) { wait = 0; }
    if (token === void 0) { token = { cancelled: false }; }
    if (wait === 0) {
        return fn;
    }
    var timeout;
    return function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        var later = function () {
            timeout = undefined;
            // check if the fn call was cancelled.
            if (!token.cancelled)
                fn.apply(void 0, args);
        };
        // because we might want to use Node.js setTimout for SSR.
        clearTimeout(timeout);
        timeout = setTimeout(later, wait);
        if (!timeout)
            fn.apply(void 0, args);
    };
};
/**
 * Normalizes the given rules expression.
 */
var normalizeRules = function (rules) {
    // if falsy value return an empty object.
    var acc = {};
    Object.defineProperty(acc, '_$$isNormalized', {
        value: true,
        writable: false,
        enumerable: false,
        configurable: false
    });
    if (!rules) {
        return acc;
    }
    // Object is already normalized, skip.
    if (isObject(rules) && rules._$$isNormalized) {
        return rules;
    }
    if (isObject(rules)) {
        return Object.keys(rules).reduce(function (prev, curr) {
            var params = [];
            if (rules[curr] === true) {
                params = [];
            }
            else if (Array.isArray(rules[curr])) {
                params = rules[curr];
            }
            else if (isObject(rules[curr])) {
                params = rules[curr];
            }
            else {
                params = [rules[curr]];
            }
            if (rules[curr] !== false) {
                prev[curr] = params;
            }
            return prev;
        }, acc);
    }
    if (typeof rules !== 'string') {
        warn('rules must be either a string or an object.');
        return acc;
    }
    return rules.split('|').reduce(function (prev, rule) {
        var parsedRule = parseRule(rule);
        if (!parsedRule.name) {
            return prev;
        }
        prev[parsedRule.name] = parsedRule.params;
        return prev;
    }, acc);
};
/**
 * Emits a warning to the console.
 */
var warn = function (message) {
    console.warn("[vee-validate] " + message);
};
/**
 * Checks if the value is an object.
 */
var isObject = function (obj) {
    return obj !== null && obj && typeof obj === 'object' && !Array.isArray(obj);
};
function identity(x) {
    return x;
}
/**
 * Checks if a function is callable.
 */
var isCallable = function (func) { return typeof func === 'function'; };
function computeClassObj(names, flags) {
    var acc = {};
    var keys = Object.keys(flags);
    var length = keys.length;
    var _loop_1 = function (i) {
        var flag = keys[i];
        var className = (names && names[flag]) || flag;
        var value = flags[flag];
        if (isNullOrUndefined(value)) {
            return "continue";
        }
        if ((flag === 'valid' || flag === 'invalid') && !flags.validated) {
            return "continue";
        }
        if (typeof className === 'string') {
            acc[className] = value;
        }
        else if (Array.isArray(className)) {
            className.forEach(function (cls) {
                acc[cls] = value;
            });
        }
    };
    for (var i = 0; i < length; i++) {
        _loop_1(i);
    }
    return acc;
}
/**
 * Converts an array-like object to array, provides a simple polyfill for Array.from
 */
function toArray(arrayLike) {
    if (isCallable(Array.from)) {
        return Array.from(arrayLike);
    }
    var array = [];
    var length = arrayLike.length;
    /* istanbul ignore next */
    for (var i = 0; i < length; i++) {
        array.push(arrayLike[i]);
    }
    /* istanbul ignore next */
    return array;
}
function findIndex(arrayLike, predicate) {
    var array = Array.isArray(arrayLike) ? arrayLike : toArray(arrayLike);
    for (var i = 0; i < array.length; i++) {
        if (predicate(array[i])) {
            return i;
        }
    }
    return -1;
}
/**
 * finds the first element that satisfies the predicate callback, polyfills array.find
 */
function find(arrayLike, predicate) {
    var array = Array.isArray(arrayLike) ? arrayLike : toArray(arrayLike);
    var idx = findIndex(array, predicate);
    return idx === -1 ? undefined : array[idx];
}
var merge = function (target, source) {
    if (!(isObject(target) && isObject(source))) {
        return target;
    }
    Object.keys(source).forEach(function (key) {
        if (isObject(source[key])) {
            if (!target[key]) {
                target[key] = {};
            }
            merge(target[key], source[key]);
            return;
        }
        target[key] = source[key];
    });
    return target;
};
function values(obj) {
    if (isCallable(Object.values)) {
        return Object.values(obj);
    }
    // fallback to keys()
    /* istanbul ignore next */
    return Object.keys(obj).map(function (k) { return obj[k]; });
}
var includes = function (collection, item) {
    return collection.indexOf(item) !== -1;
};
var isEmptyArray = function (arr) {
    return Array.isArray(arr) && arr.length === 0;
};

var LOCALE = 'en';
var Dictionary = /** @class */ (function () {
    function Dictionary(dictionary) {
        this.container = {};
        this.merge(dictionary);
    }
    Object.defineProperty(Dictionary.prototype, "locale", {
        get: function () {
            return LOCALE;
        },
        set: function (value) {
            LOCALE = value || 'en';
        },
        enumerable: true,
        configurable: true
    });
    Dictionary.prototype._hasLocale = function (locale) {
        return !!this.container[locale];
    };
    Dictionary.prototype.getMessage = function (locale, key, data) {
        var message = null;
        if (!this._hasMessage(locale, key)) {
            message = this._getDefaultMessage(locale);
        }
        else {
            message = this.container[locale].messages[key];
        }
        return isCallable(message) ? message.apply(void 0, data) : message;
    };
    /**
     * Gets a specific message for field. falls back to the rule message.
     */
    Dictionary.prototype.getFieldMessage = function (locale, field, key, data) {
        if (!this._hasLocale(locale)) {
            return this.getMessage(locale, key, data);
        }
        var dict = this.container[locale].custom && this.container[locale].custom[field];
        if (!dict || !dict[key]) {
            return this.getMessage(locale, key, data);
        }
        var message = dict[key];
        return isCallable(message) ? message.apply(void 0, data) : message;
    };
    Dictionary.prototype.merge = function (dictionary) {
        merge(this.container, dictionary);
    };
    Dictionary.prototype._getDefaultMessage = function (locale) {
        if (this._hasMessage(locale, '_default')) {
            return this.container[locale].messages._default;
        }
        return this.container.en.messages._default;
    };
    Dictionary.prototype.getAttribute = function (locale, key, fallback) {
        if (fallback === void 0) { fallback = ''; }
        if (!this._hasAttribute(locale, key)) {
            return fallback;
        }
        return this.container[locale].attributes[key];
    };
    Dictionary.prototype._hasMessage = function (locale, key) {
        return !!(this._hasLocale(locale) && this.container[locale].messages && this.container[locale].messages[key]);
    };
    Dictionary.prototype._hasAttribute = function (locale, key) {
        return !!(this._hasLocale(locale) && this.container[locale].attributes && this.container[locale].attributes[key]);
    };
    return Dictionary;
}());

var drivers = {
    default: new Dictionary({
        en: {
            messages: {},
            attributes: {},
            custom: {}
        }
    })
};
var currentDriver = 'default';
var DictionaryResolver = /** @class */ (function () {
    function DictionaryResolver() {
    }
    DictionaryResolver.setDriver = function (implementation) {
        if (typeof implementation === 'string') {
            currentDriver = implementation;
            return;
        }
        var tempName = Date.now()
            .toString()
            .substring(7);
        drivers[tempName] = implementation;
        currentDriver = tempName;
    };
    DictionaryResolver.getDriver = function () {
        return drivers[currentDriver];
    };
    return DictionaryResolver;
}());

var RULES = {};
var RuleContainer = /** @class */ (function () {
    function RuleContainer() {
    }
    RuleContainer.extend = function (name, schema) {
        // Rule already exists.
        var rule = __assign({ immediate: true, computesRequired: false }, schema);
        // rule already exists, overwrite it.
        if (RULES[name]) {
            rule = merge(RULES[name], rule);
        }
        RULES[name] = rule;
    };
    RuleContainer.isImmediate = function (name) {
        return !!(RULES[name] && RULES[name].immediate);
    };
    RuleContainer.isRequireRule = function (name) {
        return !!(RULES[name] && RULES[name].computesRequired);
    };
    RuleContainer.isTargetRule = function (name) {
        var definition = RuleContainer.getRuleDefinition(name);
        if (!definition.params) {
            return false;
        }
        return definition.params.some(function (param) { return !!param.isTarget; });
    };
    RuleContainer.getRuleDefinition = function (ruleName) {
        return RULES[ruleName];
    };
    return RuleContainer;
}());

var Validator = /** @class */ (function () {
    function Validator(options) {
        if (options === void 0) { options = { bails: true }; }
        this.bails = !!(!isNullOrUndefined(options && options.bails) ? options.bails : true);
    }
    Object.defineProperty(Validator.prototype, "locale", {
        /**
         * Getter for the current locale.
         */
        get: function () {
            return Validator.locale;
        },
        /**
         * Setter for the validator locale.
         */
        set: function (value) {
            Validator.locale = value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Validator, "locale", {
        get: function () {
            return DictionaryResolver.getDriver().locale;
        },
        /**
         * Setter for the validator locale.
         */
        set: function (value) {
            // TODO: Handle setting the locale.
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Adds and sets the current locale for the validator.
     */
    Validator.prototype.localize = function (lang, dictionary) {
        Validator.localize(lang, dictionary);
    };
    /**
     * Adds and sets the current locale for the validator.
     */
    Validator.localize = function (lang, dictionary) {
        var _a;
        if (isObject(lang)) {
            DictionaryResolver.getDriver().merge(lang);
            return;
        }
        // merge the dictionary.
        if (dictionary) {
            var locale = lang || dictionary.name;
            if (!locale)
                return;
            DictionaryResolver.getDriver().merge((_a = {},
                _a[locale] = dictionary,
                _a));
        }
        if (lang) {
            // set the locale.
            Validator.locale = lang;
        }
    };
    /**
     * Validates a value against the rules.
     */
    Validator.prototype.validate = function (value, rules, options) {
        if (options === void 0) { options = {}; }
        return __awaiter(this, void 0, void 0, function () {
            var field, result, errors, ruleMap;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        field = {
                            name: (options && options.name) || '{field}',
                            rules: normalizeRules(rules),
                            bails: getPath('bails', options, this.bails),
                            forceRequired: false,
                            crossTable: options && options.values,
                            names: (options && options.names) || {}
                        };
                        return [4 /*yield*/, this._validate(field, value, options)];
                    case 1:
                        result = _a.sent();
                        errors = [];
                        ruleMap = {};
                        result.errors.forEach(function (e) {
                            errors.push(e.msg);
                            ruleMap[e.rule] = e.msg;
                        });
                        return [2 /*return*/, {
                                valid: result.valid,
                                errors: errors,
                                failedRules: ruleMap
                            }];
                }
            });
        });
    };
    /**
     * Starts the validation process.
     */
    Validator.prototype._validate = function (field, value, _a) {
        var _b = (_a === void 0 ? {} : _a).isInitial, isInitial = _b === void 0 ? false : _b;
        return __awaiter(this, void 0, void 0, function () {
            var _c, shouldSkip, errors, rules, length, i, rule, result;
            return __generator(this, function (_d) {
                switch (_d.label) {
                    case 0: return [4 /*yield*/, this._shouldSkip(field, value)];
                    case 1:
                        _c = _d.sent(), shouldSkip = _c.shouldSkip, errors = _c.errors;
                        if (shouldSkip) {
                            return [2 /*return*/, {
                                    valid: !errors.length,
                                    errors: errors
                                }];
                        }
                        rules = Object.keys(field.rules).filter(function (rule) { return !RuleContainer.isRequireRule(rule); });
                        length = rules.length;
                        i = 0;
                        _d.label = 2;
                    case 2:
                        if (!(i < length)) return [3 /*break*/, 5];
                        if (isInitial && !RuleContainer.isImmediate(rules[i])) {
                            return [3 /*break*/, 4];
                        }
                        rule = rules[i];
                        return [4 /*yield*/, this._test(field, value, {
                                name: rule,
                                params: field.rules[rule]
                            })];
                    case 3:
                        result = _d.sent();
                        if (!result.valid) {
                            errors.push.apply(errors, result.errors);
                            if (field.bails) {
                                return [2 /*return*/, {
                                        valid: false,
                                        errors: errors
                                    }];
                            }
                        }
                        _d.label = 4;
                    case 4:
                        i++;
                        return [3 /*break*/, 2];
                    case 5: return [2 /*return*/, {
                            valid: !errors.length,
                            errors: errors
                        }];
                }
            });
        });
    };
    Validator.prototype._shouldSkip = function (field, value) {
        return __awaiter(this, void 0, void 0, function () {
            var requireRules, length, errors, isRequired, i, rule, result;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        requireRules = Object.keys(field.rules).filter(RuleContainer.isRequireRule);
                        length = requireRules.length;
                        errors = [];
                        isRequired = false;
                        i = 0;
                        _a.label = 1;
                    case 1:
                        if (!(i < length)) return [3 /*break*/, 4];
                        rule = requireRules[i];
                        return [4 /*yield*/, this._test(field, value, {
                                name: rule,
                                params: field.rules[rule]
                            })];
                    case 2:
                        result = _a.sent();
                        if (!isObject(result)) {
                            throw new Error('Require rules has to return an object (see docs)');
                        }
                        if (result.data.required) {
                            isRequired = true;
                        }
                        if (!result.valid) {
                            errors.push.apply(errors, result.errors);
                            // Exit early as the field is required and failed validation.
                            if (field.bails) {
                                return [2 /*return*/, {
                                        shouldSkip: true,
                                        errors: errors
                                    }];
                            }
                        }
                        _a.label = 3;
                    case 3:
                        i++;
                        return [3 /*break*/, 1];
                    case 4:
                        // field is configured to run through the pipeline regardless
                        if (!field.bails) {
                            return [2 /*return*/, {
                                    shouldSkip: false,
                                    errors: errors
                                }];
                        }
                        // skip if the field is not required and has an empty value.
                        return [2 /*return*/, {
                                shouldSkip: !isRequired && (isNullOrUndefined(value) || value === '' || isEmptyArray(value)),
                                errors: errors
                            }];
                }
            });
        });
    };
    /**
     * Tests a single input value against a rule.
     */
    Validator.prototype._test = function (field, value, rule) {
        return __awaiter(this, void 0, void 0, function () {
            var ruleSchema, params, result;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        ruleSchema = RuleContainer.getRuleDefinition(rule.name);
                        if (!ruleSchema.validate) {
                            throw new Error("No such validator '" + rule.name + "' exists.");
                        }
                        params = this._buildParams(rule.params, ruleSchema.params, field.crossTable);
                        return [4 /*yield*/, ruleSchema.validate(value, params)];
                    case 1:
                        result = _a.sent();
                        if (!isObject(result)) {
                            result = { valid: result, data: {} };
                        }
                        return [2 /*return*/, {
                                valid: result.valid,
                                data: result.data || {},
                                errors: result.valid ? [] : [this._createFieldError(field, rule, params, result.data)]
                            }];
                }
            });
        });
    };
    Validator.prototype._buildParams = function (provided, defined, crossTable) {
        var params = {};
        if (!defined && !Array.isArray(provided)) {
            throw new Error('You provided an object params to a rule that has no defined schema.');
        }
        // Rule probably uses an array for their args, keep it as is.
        if (Array.isArray(provided) && !defined) {
            return provided;
        }
        var definedRules;
        // collect the params schema.
        if (!defined || defined.length < provided.length) {
            var lastDefinedParam_1;
            // collect any additional parameters in the last item.
            definedRules = provided.map(function (_, idx) {
                var param = defined && defined[idx];
                lastDefinedParam_1 = param || lastDefinedParam_1;
                if (!param) {
                    param = lastDefinedParam_1;
                }
                return param;
            });
        }
        else {
            definedRules = defined;
        }
        // Match the provided array length with a temporary schema.
        for (var i = 0; i < definedRules.length; i++) {
            var options = definedRules[i];
            var value = options.default;
            // if the provided is an array, map element value.
            if (Array.isArray(provided)) {
                if (i in provided) {
                    value = provided[i];
                }
            }
            else {
                // map it from the object if it exists.
                value = options.name in provided ? provided[options.name] : value;
            }
            // if the param is a target, resolve the target value.
            if (options.isTarget) {
                value = crossTable[value];
            }
            // If there is a transformer defined.
            if (options.cast) {
                value = options.cast(value);
            }
            // already been set, probably multiple values.
            if (params[options.name]) {
                params[options.name] = Array.isArray(params[options.name]) ? params[options.name] : [params[options.name]];
                params[options.name].push(value);
            }
            else {
                // set the value.
                params[options.name] = value;
            }
        }
        return params;
    };
    /**
     * Creates a Field Error Object.
     */
    Validator.prototype._createFieldError = function (field, rule, params, data) {
        var _this = this;
        return {
            field: field.name,
            msg: this._formatErrorMessage(field, rule, params, data),
            rule: rule.name,
            regenerate: function () {
                return _this._formatErrorMessage(field, rule, params, data);
            }
        };
    };
    /**
     * Formats an error message for field and a rule.
     */
    Validator.prototype._formatErrorMessage = function (field, rule, params, data) {
        if (data === void 0) { data = {}; }
        var name = this._getFieldDisplayName(field);
        data.targetName = field.names[rule.name] || '{target}';
        data.targetName = DictionaryResolver.getDriver().getAttribute(this.locale, data.targetName) || data.targetName;
        return DictionaryResolver.getDriver().getFieldMessage(this.locale, field.name, rule.name, [name, params, data]);
    };
    /**
     * Resolves an appropriate display name, first checking 'data-as' or the registered 'prettyName'
     */
    Validator.prototype._getFieldDisplayName = function (field) {
        return field.alias || DictionaryResolver.getDriver().getAttribute(this.locale, field.name) || field.name;
    };
    return Validator;
}());

/**
 * Adds a custom validator to the list of validation rules.
 */
function extend(name, schema) {
    // makes sure new rules are properly formatted.
    guardExtend(name, schema);
    // Full schema object.
    if (typeof schema === 'object') {
        extendRule(name, schema);
        return;
    }
    extendRule(name, {
        validate: schema
    });
}
/**
 * Merges a validator object into the RULES and Messages.
 */
function extendRule(name, schema) {
    var _a, _b;
    if (schema.message) {
        DictionaryResolver.getDriver().merge((_a = {},
            _a[DictionaryResolver.getDriver().locale] = {
                messages: (_b = {},
                    _b[name] = schema.message,
                    _b)
            },
            _a));
    }
    RuleContainer.extend(name, schema);
}
/**
 * Guards from extension violations.
 */
function guardExtend(name, validator) {
    if (isCallable(validator)) {
        return;
    }
    if (isCallable(validator.validate)) {
        return;
    }
    if (RuleContainer.getRuleDefinition(name)) {
        return;
    }
    throw new Error("Extension Error: The validator '" + name + "' must be a function or have a 'validate' method.");
}

var DEFAULT_CONFIG = {
    locale: 'en',
    classNames: {
        touched: 'touched',
        untouched: 'untouched',
        valid: 'valid',
        invalid: 'invalid',
        pristine: 'pristine',
        dirty: 'dirty' // control has been interacted with
    },
    bails: true,
    aria: true,
    validity: false,
    mode: 'aggressive',
    useConstraintAttrs: true,
    dictionary: {},
    delay: 0
};
var currentConfig = __assign({}, DEFAULT_CONFIG);
var getConfig = function () { return currentConfig; };
var setConfig = function (newConf) {
    currentConfig = __assign({}, currentConfig, newConf);
};

var isEvent = function (evt) {
    if (!evt) {
        return false;
    }
    // tslint:disable-next-line
    if (typeof Event !== 'undefined' && isCallable(Event) && evt instanceof Event) {
        return true;
    }
    if (evt && evt.srcElement) {
        return true;
    }
    return false;
};
var normalizeEvents = function (evts) {
    if (!evts)
        return [];
    return typeof evts === 'string' ? evts.split('|') : evts;
};
function normalizeEventValue(value) {
    if (!isEvent(value)) {
        return value;
    }
    var input = value.target;
    if (input.type === 'file' && input.files) {
        return toArray(input.files);
    }
    return input.value;
}

var isTextInput = function (vnode) {
    var attrs = (vnode.data && vnode.data.attrs) || vnode.elm;
    return includes(['text', 'password', 'search', 'email', 'tel', 'url', 'textarea', 'number'], attrs && attrs.type);
};
// Gets the model object on the vnode.
function findModel(vnode) {
    if (!vnode.data) {
        return null;
    }
    // Component Model
    // THIS IS NOT TYPED IN OFFICIAL VUE TYPINGS
    var nonStandardVNodeData = vnode.data;
    if ('model' in nonStandardVNodeData) {
        return nonStandardVNodeData.model;
    }
    if (!vnode.data.directives) {
        return null;
    }
    return find(vnode.data.directives, function (d) { return d.name === 'model'; });
}
function extractChildren(vnode) {
    if (Array.isArray(vnode)) {
        return vnode;
    }
    if (Array.isArray(vnode.children)) {
        return vnode.children;
    }
    if (vnode.componentOptions && Array.isArray(vnode.componentOptions.children)) {
        return vnode.componentOptions.children;
    }
    return [];
}
function extractVNodes(vnode) {
    if (!Array.isArray(vnode) && findModel(vnode)) {
        return [vnode];
    }
    var children = extractChildren(vnode);
    return children.reduce(function (nodes, node) {
        var candidates = extractVNodes(node);
        if (candidates.length) {
            nodes.push.apply(nodes, candidates);
        }
        return nodes;
    }, []);
}
// Resolves v-model config if exists.
function findModelConfig(vnode) {
    if (!vnode.componentOptions)
        return null;
    // This is also not typed in the standard Vue TS.
    return vnode.componentOptions.Ctor.options.model;
}
// Adds a listener to vnode listener object.
function mergeVNodeListeners(obj, eventName, handler) {
    // Has a single listener.
    if (isCallable(obj[eventName])) {
        var prevHandler = obj[eventName];
        obj[eventName] = [prevHandler];
    }
    // has other listeners.
    if (Array.isArray(obj[eventName])) {
        obj[eventName].push(handler);
        return;
    }
    // no listener at all.
    if (isNullOrUndefined(obj[eventName])) {
        obj[eventName] = [handler];
    }
}
// Adds a listener to a native HTML vnode.
function addNativeNodeListener(node, eventName, handler) {
    if (!node.data) {
        node.data = {};
    }
    if (isNullOrUndefined(node.data.on)) {
        node.data.on = {};
    }
    mergeVNodeListeners(node.data.on, eventName, handler);
}
// Adds a listener to a Vue component vnode.
function addComponentNodeListener(node, eventName, handler) {
    if (!node.componentOptions) {
        return;
    }
    /* istanbul ignore next */
    if (!node.componentOptions.listeners) {
        node.componentOptions.listeners = {};
    }
    mergeVNodeListeners(node.componentOptions.listeners, eventName, handler);
}
function addVNodeListener(vnode, eventName, handler) {
    if (vnode.componentOptions) {
        addComponentNodeListener(vnode, eventName, handler);
    }
    addNativeNodeListener(vnode, eventName, handler);
}
// Determines if `change` should be used over `input` for listeners.
function getInputEventName(vnode, model) {
    // Is a component.
    if (vnode.componentOptions) {
        var event_1 = (findModelConfig(vnode) || { event: 'input' }).event;
        return event_1;
    }
    // Lazy Models typically use change event
    if (model && model.modifiers && model.modifiers.lazy) {
        return 'change';
    }
    // is a textual-type input.
    if (vnode.data && vnode.data.attrs && isTextInput(vnode)) {
        return 'input';
    }
    return 'change';
}
// TODO: Type this one properly.
function normalizeSlots(slots, ctx) {
    return Object.keys(slots).reduce(function (arr, key) {
        slots[key].forEach(function (vnode) {
            if (!vnode.context) {
                slots[key].context = ctx;
                if (!vnode.data) {
                    vnode.data = {};
                }
                vnode.data.slot = key;
            }
        });
        return arr.concat(slots[key]);
    }, []);
}
function resolveTextualRules(vnode) {
    var attrs = vnode.data && vnode.data.attrs;
    var rules = {};
    if (!attrs)
        return rules;
    if (attrs.type === 'email') {
        rules.email = ['multiple' in attrs];
    }
    if (attrs.pattern) {
        rules.regex = attrs.pattern;
    }
    if (attrs.maxlength >= 0) {
        rules.max = attrs.maxlength;
    }
    if (attrs.minlength >= 0) {
        rules.min = attrs.minlength;
    }
    if (attrs.type === 'number') {
        rules.decimal = [];
        if (attrs.min !== '') {
            rules.min_value = Number(attrs.min);
        }
        if (attrs.max !== '') {
            rules.max_value = Number(attrs.max);
        }
    }
    return rules;
}
function resolveRules(vnode) {
    var htmlTags = ['input', 'select'];
    var attrs = vnode.data && vnode.data.attrs;
    if (!includes(htmlTags, vnode.tag) || !attrs) {
        return {};
    }
    var rules = {};
    if ('required' in attrs) {
        rules.required = attrs.type === 'checkbox' ? [true] : true;
    }
    if (isTextInput(vnode)) {
        return normalizeRules(__assign({}, rules, resolveTextualRules(vnode)));
    }
    return normalizeRules(rules);
}
function normalizeChildren(context, slotProps) {
    if (context.$scopedSlots.default) {
        return context.$scopedSlots.default(slotProps) || [];
    }
    return context.$slots.default || [];
}

var aggressive = function () { return ({
    on: ['input']
}); };
var lazy = function () { return ({
    on: ['change']
}); };
var eager = function (_a) {
    var errors = _a.errors;
    if (errors.length) {
        return {
            on: ['input']
        };
    }
    return {
        on: ['change', 'blur']
    };
};
var passive = function () { return ({
    on: []
}); };
var modes = {
    aggressive: aggressive,
    eager: eager,
    passive: passive,
    lazy: lazy
};

/**
 * Determines if a provider needs to run validation.
 */
function shouldValidate(ctx, model) {
    // when an immediate/initial validation is needed and wasn't done before.
    if (!ctx._ignoreImmediate && ctx.immediate) {
        return true;
    }
    // when the value changes for whatever reason.
    if (ctx.value !== model.value) {
        return true;
    }
    // when it needs validation due to props/cross-fields changes.
    if (ctx._needsValidation) {
        return true;
    }
    // when the initial value is undefined and the field wasn't rendered yet.
    if (!ctx.initialized && model.value === undefined) {
        return true;
    }
    return false;
}
function createValidationCtx(ctx) {
    return {
        errors: ctx.messages,
        flags: ctx.flags,
        classes: ctx.classes,
        valid: ctx.isValid,
        failedRules: ctx.failedRules,
        reset: function () { return ctx.reset(); },
        validate: function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return ctx.validate.apply(ctx, args);
        },
        aria: {
            'aria-invalid': ctx.flags.invalid ? 'true' : 'false',
            'aria-required': ctx.isRequired ? 'true' : 'false'
        }
    };
}
function onRenderUpdate(vm, model) {
    if (!vm.initialized) {
        vm.initialValue = model.value;
    }
    var validateNow = shouldValidate(vm, model);
    vm._needsValidation = false;
    vm.value = model.value;
    vm._ignoreImmediate = true;
    if (!validateNow) {
        return;
    }
    vm.validateSilent().then(vm.immediate || vm.flags.validated ? vm.applyResult : function (x) { return x; });
}
function computeModeSetting(ctx) {
    var compute = isCallable(ctx.mode) ? ctx.mode : modes[ctx.mode];
    return compute({
        errors: ctx.messages,
        value: ctx.value,
        flags: ctx.flags
    });
}
// Creates the common handlers for a validatable context.
function createCommonHandlers(vm) {
    var onInput = function (e) {
        vm.syncValue(e); // track and keep the value updated.
        vm.setFlags({ dirty: true, pristine: false });
    };
    // Blur event listener.
    var onBlur = function () {
        vm.setFlags({ touched: true, untouched: false });
    };
    var onValidate = vm.$veeHandler;
    var mode = computeModeSetting(vm);
    // Handle debounce changes.
    if (!onValidate || vm.$veeDebounce !== vm.debounce) {
        onValidate = debounce(function () {
            vm.$nextTick(function () {
                var pendingPromise = vm.validateSilent();
                // avoids race conditions between successive validations.
                vm._pendingValidation = pendingPromise;
                // tslint:disable-next-line
                pendingPromise.then(function (result) {
                    if (pendingPromise === vm._pendingValidation) {
                        vm.applyResult(result);
                        vm._pendingValidation = null;
                    }
                });
            });
        }, mode.debounce || vm.debounce);
        // Cache the handler so we don't create it each time.
        vm.$veeHandler = onValidate;
        // cache the debounce value so we detect if it was changed.
        vm.$veeDebounce = vm.debounce;
    }
    return { onInput: onInput, onBlur: onBlur, onValidate: onValidate };
}
// Adds all plugin listeners to the vnode.
function addListeners(vm, node) {
    var model = findModel(node);
    // cache the input eventName.
    vm._inputEventName = vm._inputEventName || getInputEventName(node, model);
    onRenderUpdate(vm, model);
    var _a = createCommonHandlers(vm), onInput = _a.onInput, onBlur = _a.onBlur, onValidate = _a.onValidate;
    addVNodeListener(node, vm._inputEventName, onInput);
    addVNodeListener(node, 'blur', onBlur);
    // add the validation listeners.
    vm.normalizedEvents.forEach(function (evt) {
        addVNodeListener(node, evt, onValidate);
    });
    vm.initialized = true;
}

var VALIDATOR;
var PROVIDER_COUNTER = 0;
function data() {
    var messages = [];
    var defaultValues = {
        messages: messages,
        value: undefined,
        initialized: false,
        initialValue: undefined,
        flags: createFlags(),
        failedRules: {},
        forceRequired: false,
        isDeactivated: false,
        id: ''
    };
    return defaultValues;
}
var ValidationProvider = Vue.extend({
    inject: {
        $_veeObserver: {
            from: '$_veeObserver',
            default: function () {
                if (!this.$vnode.context.$_veeObserver) {
                    this.$vnode.context.$_veeObserver = createObserver();
                }
                return this.$vnode.context.$_veeObserver;
            }
        }
    },
    props: {
        vid: {
            type: String,
            default: function () {
                PROVIDER_COUNTER++;
                return "_vee_" + PROVIDER_COUNTER;
            }
        },
        name: {
            type: String,
            default: null
        },
        mode: {
            type: [String, Function],
            default: function () {
                return getConfig().mode;
            }
        },
        rules: {
            type: [Object, String],
            default: null
        },
        immediate: {
            type: Boolean,
            default: false
        },
        persist: {
            type: Boolean,
            default: false
        },
        bails: {
            type: Boolean,
            default: function () { return getConfig().bails; }
        },
        debounce: {
            type: Number,
            default: function () { return getConfig().delay || 0; }
        },
        tag: {
            type: String,
            default: 'span'
        },
        slim: {
            type: Boolean,
            default: false
        }
    },
    watch: {
        rules: {
            deep: true,
            handler: function (val, oldVal) {
                this._needsValidation = !isEqual(val, oldVal);
            }
        }
    },
    data: data,
    computed: {
        isValid: function () {
            return this.flags.valid;
        },
        fieldDeps: function () {
            var _this = this;
            return Object.keys(this.normalizedRules)
                .filter(RuleContainer.isTargetRule)
                .map(function (rule) {
                var depName = _this.normalizedRules[rule][0];
                watchCrossFieldDep(_this, depName);
                return depName;
            });
        },
        normalizedEvents: function () {
            var _this = this;
            var on = computeModeSetting(this).on;
            return normalizeEvents(on || []).map(function (e) {
                if (e === 'input') {
                    return _this._inputEventName;
                }
                return e;
            });
        },
        isRequired: function () {
            var rules = __assign({}, this._resolvedRules, this.normalizedRules);
            var forceRequired = this.forceRequired;
            var isRequired = rules.required || forceRequired;
            this.flags.required = isRequired;
            return isRequired;
        },
        classes: function () {
            var names = getConfig().classNames;
            return computeClassObj(names, this.flags);
        },
        normalizedRules: function () {
            return normalizeRules(this.rules);
        }
    },
    render: function (h) {
        var _this = this;
        this.registerField();
        var ctx = createValidationCtx(this);
        var children = normalizeChildren(this, ctx);
        // Handle single-root slot.
        extractVNodes(children).forEach(function (input) {
            _this._resolvedRules = resolveRules(input);
            addListeners(_this, input);
        });
        return this.slim && children.length <= 1 ? children[0] : h(this.tag, children);
    },
    beforeDestroy: function () {
        // cleanup reference.
        this.$_veeObserver.unsubscribe(this);
    },
    activated: function () {
        this.$_veeObserver.subscribe(this);
        this.isDeactivated = false;
    },
    deactivated: function () {
        this.$_veeObserver.unsubscribe(this);
        this.isDeactivated = true;
    },
    methods: {
        setFlags: function (flags) {
            var _this = this;
            Object.keys(flags).forEach(function (flag) {
                _this.flags[flag] = flags[flag];
            });
        },
        syncValue: function (v) {
            var value = normalizeEventValue(v);
            this.value = value;
            this.flags.changed = this.initialValue !== value;
        },
        reset: function () {
            this.messages = [];
            this.initialValue = this.value;
            var flags = createFlags();
            this.setFlags(flags);
        },
        validate: function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return __awaiter(this, void 0, void 0, function () {
                var result;
                return __generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            if (args.length > 0) {
                                this.syncValue(args[0]);
                            }
                            return [4 /*yield*/, this.validateSilent()];
                        case 1:
                            result = _a.sent();
                            this.applyResult(result);
                            return [2 /*return*/, result];
                    }
                });
            });
        },
        validateSilent: function () {
            return __awaiter(this, void 0, void 0, function () {
                var rules, result;
                return __generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            this.setFlags({ pending: true });
                            rules = __assign({}, this._resolvedRules, this.normalizedRules);
                            Object.defineProperty(rules, '_$$isNormalized', {
                                value: true,
                                writable: false,
                                enumerable: false,
                                configurable: false
                            });
                            return [4 /*yield*/, VALIDATOR.validate(this.value, rules, {
                                    name: this.name,
                                    values: createValuesLookup(this),
                                    bails: this.bails,
                                    isInitial: !this.initialized
                                })];
                        case 1:
                            result = _a.sent();
                            this.setFlags({ pending: false });
                            if (!this.isRequired) {
                                this.setFlags({ valid: result.valid, invalid: !result.valid });
                            }
                            return [2 /*return*/, result];
                    }
                });
            });
        },
        applyResult: function (_a) {
            var errors = _a.errors, failedRules = _a.failedRules;
            this.messages = errors;
            this.failedRules = __assign({}, failedRules);
            this.setFlags({
                valid: !errors.length,
                changed: this.value !== this.initialValue,
                invalid: !!errors.length,
                validated: true
            });
        },
        registerField: function () {
            if (!VALIDATOR) {
                VALIDATOR = new Validator({ bails: getConfig().bails });
            }
            updateRenderingContextRefs(this);
        }
    }
});
function createValuesLookup(vm) {
    var providers = vm.$_veeObserver.refs;
    var reduced = {};
    return vm.fieldDeps.reduce(function (acc, depName) {
        if (!providers[depName]) {
            return acc;
        }
        acc[depName] = providers[depName].value;
        return acc;
    }, reduced);
}
function updateRenderingContextRefs(vm) {
    // IDs should not be nullable.
    if (isNullOrUndefined(vm.id) && vm.id === vm.vid) {
        vm.id = "" + PROVIDER_COUNTER;
        PROVIDER_COUNTER++;
    }
    var id = vm.id, vid = vm.vid;
    // Nothing has changed.
    if (vm.isDeactivated || (id === vid && vm.$_veeObserver.refs[id])) {
        return;
    }
    // vid was changed.
    if (id !== vid && vm.$_veeObserver.refs[id] === vm) {
        vm.$_veeObserver.unsubscribe(vm);
    }
    vm.$_veeObserver.subscribe(vm);
    vm.id = vid;
}
function createObserver() {
    return {
        refs: {},
        subscribe: function (ctx) {
            this.refs[ctx.vid] = ctx;
        },
        unsubscribe: function (ctx) {
            delete this.refs[ctx.vid];
        }
    };
}
function watchCrossFieldDep(ctx, depName, withHooks) {
    if (withHooks === void 0) { withHooks = true; }
    var providers = ctx.$_veeObserver.refs;
    if (!ctx._veeWatchers) {
        ctx._veeWatchers = {};
    }
    if (!providers[depName] && withHooks) {
        return ctx.$once('hook:mounted', function () {
            watchCrossFieldDep(ctx, depName, false);
        });
    }
    if (!isCallable(ctx._veeWatchers[depName]) && providers[depName]) {
        ctx._veeWatchers[depName] = providers[depName].$watch('value', function () {
            if (ctx.flags.validated) {
                ctx._needsValidation = true;
                // tslint:disable-next-line: no-floating-promises
                ctx.validate();
            }
        });
    }
}

var flagMergingStrategy = {
    pristine: 'every',
    dirty: 'some',
    touched: 'some',
    untouched: 'every',
    valid: 'every',
    invalid: 'some',
    pending: 'some',
    validated: 'every'
};
function mergeFlags(lhs, rhs, strategy) {
    var stratName = flagMergingStrategy[strategy];
    return [lhs, rhs][stratName](function (f) { return f; });
}
var OBSERVER_COUNTER = 0;
function data$1() {
    var refs = {};
    var inactiveRefs = {};
    // FIXME: Not sure of this one can be typed, circular type reference.
    var observers = [];
    return {
        refs: refs,
        observers: observers,
        inactiveRefs: inactiveRefs
    };
}
var ValidationObserver = Vue.extend({
    name: 'ValidationObserver',
    provide: function () {
        return {
            $_veeObserver: this
        };
    },
    inject: {
        $_veeObserver: {
            from: '$_veeObserver',
            default: function () {
                if (!this.$vnode.context.$_veeObserver) {
                    return null;
                }
                return this.$vnode.context.$_veeObserver;
            }
        }
    },
    props: {
        tag: {
            type: String,
            default: 'span'
        },
        vid: {
            type: String,
            default: function () {
                return "obs_" + OBSERVER_COUNTER++;
            }
        },
        slim: {
            type: Boolean,
            default: false
        }
    },
    data: data$1,
    computed: {
        ctx: function () {
            var _this = this;
            var ctx = {
                errors: {},
                validate: function (arg) {
                    var promise = _this.validate(arg);
                    return {
                        then: function (thenable) {
                            // tslint:disable-next-line
                            promise.then(function (success) {
                                if (success && isCallable(thenable)) {
                                    return Promise.resolve(thenable());
                                }
                                return Promise.resolve(success);
                            });
                        }
                    };
                },
                reset: function () { return _this.reset(); }
            };
            return values(this.refs).concat(Object.keys(this.inactiveRefs).map(function (key) {
                return {
                    vid: key,
                    flags: _this.inactiveRefs[key].flags,
                    messages: _this.inactiveRefs[key].errors
                };
            }), this.observers).reduce(function (acc, provider) {
                Object.keys(flagMergingStrategy).forEach(function (flag) {
                    var flags = provider.flags || provider.ctx;
                    if (!(flag in acc)) {
                        acc[flag] = flags[flag];
                        return;
                    }
                    acc[flag] = mergeFlags(acc[flag], flags[flag], flag);
                });
                acc.errors[provider.vid] =
                    provider.messages ||
                        values(provider.ctx.errors).reduce(function (errs, obsErrors) {
                            return errs.concat(obsErrors);
                        }, []);
                return acc;
            }, ctx);
        }
    },
    created: function () {
        if (this.$_veeObserver) {
            this.$_veeObserver.subscribe(this, 'observer');
        }
    },
    activated: function () {
        if (this.$_veeObserver) {
            this.$_veeObserver.subscribe(this, 'observer');
        }
    },
    deactivated: function () {
        if (this.$_veeObserver) {
            this.$_veeObserver.unsubscribe(this, 'observer');
        }
    },
    beforeDestroy: function () {
        if (this.$_veeObserver) {
            this.$_veeObserver.unsubscribe(this, 'observer');
        }
    },
    render: function (h) {
        var children = normalizeChildren(this, this.ctx);
        return this.slim && children.length <= 1 ? children[0] : h(this.tag, children);
    },
    methods: {
        subscribe: function (subscriber, kind) {
            var _a;
            if (kind === void 0) { kind = 'provider'; }
            if (kind === 'observer') {
                this.observers.push(subscriber);
                return;
            }
            this.refs = Object.assign({}, this.refs, (_a = {}, _a[subscriber.vid] = subscriber, _a));
            if (subscriber.persist && this.inactiveRefs[subscriber.vid]) {
                this.restoreProviderState(subscriber);
            }
        },
        unsubscribe: function (_a, kind) {
            var vid = _a.vid;
            if (kind === void 0) { kind = 'provider'; }
            if (kind === 'provider') {
                this.removeProvider(vid);
            }
            var idx = findIndex(this.observers, function (o) { return o.vid === vid; });
            if (idx !== -1) {
                this.observers.splice(idx, 1);
            }
        },
        validate: function (_a) {
            var _b = (_a === void 0 ? {} : _a).silent, silent = _b === void 0 ? false : _b;
            return __awaiter(this, void 0, void 0, function () {
                var results;
                return __generator(this, function (_c) {
                    switch (_c.label) {
                        case 0: return [4 /*yield*/, Promise.all(values(this.refs).map(function (ref) {
                                return ref[silent ? 'validateSilent' : 'validate']().then(function (r) { return r.valid; });
                            }).concat(this.observers.map(function (obs) { return obs.validate({ silent: silent }); })))];
                        case 1:
                            results = _c.sent();
                            return [2 /*return*/, results.every(function (r) { return r; })];
                    }
                });
            });
        },
        reset: function () {
            var _this = this;
            Object.keys(this.inactiveRefs).forEach(function (key) {
                _this.$delete(_this.inactiveRefs, key);
            });
            return values(this.refs).concat(this.observers).forEach(function (ref) { return ref.reset(); });
        },
        restoreProviderState: function (provider) {
            var state = this.inactiveRefs[provider.vid];
            provider.setFlags(state.flags);
            provider.applyResult(state);
            this.$delete(this.inactiveRefs, provider.vid);
        },
        removeProvider: function (vid) {
            var provider = this.refs[vid];
            // save it for the next time.
            if (provider && provider.persist) {
                /* istanbul ignore next */
                if (process.env.NODE_ENV !== 'production') {
                    if (vid.indexOf('_vee_') === 0) {
                        warn('Please provide a `vid` prop when using `persist`, there might be unexpected issues otherwise.');
                    }
                }
                this.inactiveRefs[vid] = {
                    flags: provider.flags,
                    errors: provider.messages,
                    failedRules: provider.failedRules
                };
            }
            this.$delete(this.refs, vid);
        }
    }
});

function withValidation(component, mapProps) {
    if (mapProps === void 0) { mapProps = identity; }
    var options = 'options' in component ? component.options : component;
    var providerOpts = ValidationProvider.options;
    var hoc = {
        name: (options.name || 'AnonymousHoc') + "WithValidation",
        props: __assign({}, providerOpts.props),
        data: providerOpts.data,
        computed: __assign({}, providerOpts.computed),
        methods: __assign({}, providerOpts.methods),
        beforeDestroy: providerOpts.beforeDestroy,
        inject: providerOpts.inject
    };
    var eventName = (options.model && options.model.event) || 'input';
    hoc.render = function (h) {
        var _a;
        this.registerField();
        var vctx = createValidationCtx(this);
        var listeners = __assign({}, this.$listeners);
        var model = findModel(this.$vnode);
        this._inputEventName = this._inputEventName || getInputEventName(this.$vnode, model);
        onRenderUpdate(this, model);
        var _b = createCommonHandlers(this), onInput = _b.onInput, onBlur = _b.onBlur, onValidate = _b.onValidate;
        mergeVNodeListeners(listeners, eventName, onInput);
        mergeVNodeListeners(listeners, 'blur', onBlur);
        this.normalizedEvents.forEach(function (evt) {
            mergeVNodeListeners(listeners, evt, onValidate);
        });
        // Props are any attrs not associated with ValidationProvider Plus the model prop.
        // WARNING: Accidental prop overwrite will probably happen.
        var prop = (findModelConfig(this.$vnode) || { prop: 'value' }).prop;
        var props = __assign({}, this.$attrs, (_a = {}, _a[prop] = model.value, _a), mapProps(vctx));
        return h(options, {
            attrs: this.$attrs,
            props: props,
            on: listeners
        }, normalizeSlots(this.$slots, this.$vnode.context));
    };
    return hoc;
}

var configure = function (cfg) {
    setConfig(cfg);
};
var setI18nDriver = function (driver) {
    DictionaryResolver.setDriver(driver);
};
var setMode = function (mode, implementation) {
    setConfig({ mode: mode });
    if (!implementation) {
        return;
    }
    if (!isCallable(implementation)) {
        throw new Error('A mode implementation must be a function');
    }
    modes[mode] = implementation;
};

var version = '3.0.0-beta.0';

export { ValidationObserver, ValidationProvider, Validator, configure, extend, setI18nDriver, setMode, version, withValidation };
